@@ -18,9 +18,7 @@ REQUIREMENTS_IN = [  # this is actually ordered, as files depend on each other
 ]
 
 SUPPORTED = [
-    "3.9",
     "3.10",
-    "pypy3.10",
     "3.11",
     "pypy3.11",
     "3.12",
@@ -86,15 +84,6 @@ def tests(session):
         session.run("pytest", "--parallel-threads=10", *session.posargs, TESTS)
 
 
-@session()
-def audit(session):
-    """
-    Audit dependencies for vulnerabilities.
-    """
-    session.install("pip-audit", ROOT)
-    session.run("python", "-m", "pip_audit")
-
-
 @session(tags=["build"])
 def build(session):
     """
